#!/usr/bin/env python

import os
import sys
import curses
import logging
import argparse

from foundation.utils import Workers, WORKER_NAME, HostWorker
from foundation.workers import list_workers, select_worker

logging.basicConfig(level=logging.CRITICAL)

parser = argparse.ArgumentParser(description="Start an HCI worker.")
parser.add_argument(
    '-a', '--advertise_addr', default=None, help="Advertise address."
)
args = parser.parse_args()

workers = Workers(swarm_advertise_addr=args.advertise_addr)
host_workers = HostWorker()

foundation_services = set(
    [
        # 'kafka-service',
        # 'zookeeper-service',
        # 'kafka-logs-service',
        # 'zookeeper-logs-service',
        'timescaledb-service',
        'jupyterlab-service',
        'ntp-service',
    ]
)

COMMANDS = [
    "['q' to close]",
    "['k' to kill all services]",
    "['l' to kill all workers]",
    # "['s' to run Foundation]",
]
COMMAND_START = '[START]'
COMMAND_RESTART = '[RESTART]'
COMMAND_STOP = '[STOP]'
COMMAND_KILL = '[KILL]'
COMMAND_TERMINATE = '[TERM]'


system_workers = set(
    [
        WORKER_NAME.format(worker).replace('_', '-')
        for worker in list_workers()
    ]
)

extra_args = {
    'start_timescaledb_api_worker': {
        'service_name': 'timescaledb_api',
        'image': 'djangorun',
        'endpoint': '/timescaledbapp/',
        'port': 51102,
        'restart': True,
    },
    'start_chaski_root_worker': {
        'port': 51110,
    },
    'start_chaski_ca_worker': {
        'port': 51111,
    },
    'start_chaski_remote_worker': {
        'port': 51112,
    },
    'start_chaski2api_worker': {
        'port': 51113,
    },
    'start_chaski_logger_root_worker': {
        'port': 51114,
    },
    'start_chaski_api_logger_worker': {
        'port': 51115,
    },
}

CONTINUE = True


########################################################################
class Stats:
    """"""

    COL = [0, 16, 60, 73, 87, 109, 126, 164]
    COL_NAMES = [
        "",
        "RUNNING",
        "NODE",
        "SERVICE ID",
        "CONTAINER ID",
        "IMAGE",
        "URL",
    ]

    # ----------------------------------------------------------------------
    def __init__(self, stdscr):
        """Constructor"""
        global CONTINUE
        self.row_services = {}
        self.row_host_workers = {}
        curses.resizeterm(100, 1000)
        curses.curs_set(0)
        stdscr.nodelay(True)
        curses.mousemask(1)
        self.stdscr = stdscr

        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(4, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(5, curses.COLOR_CYAN, curses.COLOR_BLACK)

        while CONTINUE:
            self.stdscr.clear()
            self.stdscr.addstr(0, 0, '-' * 80)
            start = self.update_services_stats(
                foundation_services, name='FOUNDATION SERVICES', start=1
            )
            start = self.update_services_stats(
                system_workers, name='SYSTEM WORKERS', start=start
            )
            start = self.update_services_stats(
                set(workers.swarm.services)
                - foundation_services
                - system_workers,
                name='USER WORKERS',
                start=start,
            )

            start = self.update_host_workers('HOST WORKERS', start=start)

            self.stdscr.addstr(start, 0, '-' * 80)
            start = self.update_volumes_stats(
                name='VOLUMES', start=start + 2
            )

            self.stdscr.addstr(start, 0, '-' * 80)
            self.stdscr.addstr(
                start + 2, 0, workers.swarm.get_join_command()
            )
            self.stdscr.addstr(
                start + 10, 0, ' '.join(COMMANDS), curses.color_pair(5)
            )
            self.stdscr.addstr(start + 4, 0, '')  # logs
            self.stdscr.refresh()

            event = stdscr.getch()
            if event == curses.KEY_MOUSE:
                _, x, y, _, _ = curses.getmouse()
                command = stdscr.instr(
                    y, max([0, x - 7]), 15 + min([0, x - 7])
                ).decode('utf-8')
                line = stdscr.instr(y, 0, 500).decode('utf-8')
                if len(line.split()) > 1:
                    worker = line.split()[1][line.split()[1].find('@') + 1 :]
                else:
                    worker = None
                self.process_service_event(command, y, worker)
                self.process_host_event(command, y, worker)

            elif event == ord('q'):
                CONTINUE = False
                break

            elif event == ord('k'):
                workers.swarm.stop_all_services()

            elif event == ord('l'):
                workers.stop_all_workers()

            # elif event == ord('s'):
            #     os.system('foundation_start > /dev/null 2>&1')

    # ----------------------------------------------------------------------
    def process_service_event(self, command, row, worker):
        """"""
        if not row in self.row_services:
            return

        if COMMAND_RESTART in command:
            command = COMMAND_RESTART
        elif COMMAND_STOP in command:
            command = COMMAND_STOP
        elif COMMAND_START in command:
            command = COMMAND_START

        service_name = self.row_services[row]
        if command == COMMAND_STOP:
            workers.swarm.stop_service(service_name)
        elif command in [COMMAND_RESTART, COMMAND_START]:

            method_service = f'start_{service_name.replace("-service",  "").replace("-",  "_")}'
            # service_name = service_name.replace("-worker", "").replace("-", "_")
            worker_name = service_name.replace("-worker", "").replace(
                "-", "_"
            )

            if hasattr(workers.swarm, method_service):
                getattr(workers.swarm, method_service)(
                    **extra_args.get(
                        method_service,
                        {
                            'restart': True,
                        },
                    )
                )

            elif worker_name in workers.swarm.services or os.path.exists(
                select_worker(worker_name)
            ):
                workers.start_worker(
                    worker_name,
                    **extra_args.get(
                        method_service,
                        {
                            'restart': True,
                        },
                    ),
                )

            elif worker_name in workers.swarm.services:
                workers.start_worker(
                    worker,
                    service_name=worker_name,
                    **extra_args.get(
                        method_service,
                        {
                            'restart': True,
                        },
                    ),
                )

    # ----------------------------------------------------------------------
    def process_host_event(self, command, row, service_name):
        """"""
        if not row in self.row_host_workers:
            return

        if COMMAND_RESTART in command:
            command = COMMAND_RESTART
        elif COMMAND_STOP in command:
            command = COMMAND_STOP
        elif COMMAND_START in command:
            command = COMMAND_START
        elif COMMAND_TERMINATE in command:
            command = COMMAND_TERMINATE
        elif COMMAND_KILL in command:
            command = COMMAND_KILL

        # host_workers.load_ids()

        if command == COMMAND_KILL:
            host_workers.stop(service_name, 'kill')
        elif command == COMMAND_TERMINATE:
            host_workers.stop(service_name, 'terminate')
        elif command in [COMMAND_RESTART, COMMAND_START]:
            host_workers.restart(service_name)

    # ----------------------------------------------------------------------
    def write_row(self, items, row=0, m=0):
        """"""
        for item, col in zip(items, self.COL):
            if ('----' in items[:-1] or 'N/A' in items[:-1]) and (
                col == self.COL[1]
            ):
                color = curses.color_pair(1)
            elif (items[2] == 'True') and (col == self.COL[1]):
                color = curses.color_pair(2)
            elif col == 0 and item in [
                COMMAND_START,
                COMMAND_STOP + COMMAND_RESTART,
                COMMAND_KILL + COMMAND_RESTART,
                COMMAND_TERMINATE + COMMAND_RESTART,
            ]:
                if item == COMMAND_START:
                    color = curses.color_pair(3)
                elif item in [
                    (COMMAND_STOP + COMMAND_RESTART),
                    (COMMAND_KILL + COMMAND_RESTART),
                    (COMMAND_TERMINATE + COMMAND_RESTART),
                ]:
                    color = curses.color_pair(4)
                else:
                    color = curses.color_pair(0)
            elif col == self.COL[1] and items[3] in ['None', 'zombie']:
                color = curses.color_pair(1)
            elif col == 126 and not item in ['IMAGE', '----']:
                images = workers.swarm.client.images.list()
                if not any(
                    item in tag for image in images for tag in image.tags
                ):
                    color = curses.color_pair(1)
            else:
                color = curses.color_pair(0)

            self.stdscr.addstr(row, col, item, color)

    # ----------------------------------------------------------------------
    def service_get(self, service, attr='id', default='----'):
        """"""
        service = workers.swarm.client.services.list(
            filters={
                'name': service,
            }
        )
        if service:
            try:
                match attr:
                    case 'id':
                        return service[0].id[:12]
                    case 'container_id':
                        return service[0].tasks()[0]['Status'][
                            'ContainerStatus'
                        ]['ContainerID'][:12]
                    case 'image':
                        return service[0].tasks()[0]['Spec'][
                            'ContainerSpec'
                        ]['Image']
                    case 'ip':
                        return list(
                            filter(
                                lambda s: s.startswith('PORT='),
                                service[0].tasks()[0]['Spec'][
                                    'ContainerSpec'
                                ]['Env'],
                            )
                        )[0].replace('PORT=', '')
                    case 'url':
                        return list(
                            filter(
                                lambda s: s.startswith('URL='),
                                service[0].tasks()[0]['Spec'][
                                    'ContainerSpec'
                                ]['Env'],
                            )
                        )[0].replace('URL=', '')
                    case 'service_name':
                        try:
                            return "@" + list(
                                filter(
                                    lambda s: s.startswith('SERVICE_NAME='),
                                    service[0].tasks()[0]['Spec'][
                                        'ContainerSpec'
                                    ]['Env'],
                                )
                            )[0].replace('SERVICE_NAME=', '')
                        except:
                            return ''
                    case 'endpoint':
                        try:
                            return list(
                                filter(
                                    lambda s: s.startswith('ENDPOINT='),
                                    service[0].tasks()[0]['Spec'][
                                        'ContainerSpec'
                                    ]['Env'],
                                )
                            )[0].replace('ENDPOINT=', '')
                        except:
                            return ''
                    case 'node':

                        node_id = service[0].tasks()[0]['NodeID']

                        node_info = workers.swarm.client.nodes.get(node_id)

                        return node_info.attrs['Description']['Hostname']

            except:
                return "N/A"
        else:
            return default

    # ----------------------------------------------------------------------
    def update_services_stats(self, services, name, start=0):
        """"""
        if not services:
            return start
        self.write_row([name] + self.COL_NAMES, row=start, m=0)
        for row, service in enumerate(services, start=start + 1):

            self.row_services[row] = service

            url = '----'
            if self.service_get(service, attr='ip') not in ['----', 'N/A']:
                url = f"http://127.0.0.1:{self.service_get(service, attr='ip')}{self.service_get(service, attr='endpoint')}"
            elif self.service_get(service, attr='url') not in [
                '----',
                'N/A',
            ]:
                url = f"{self.service_get(service, attr='url')}:{self.service_get(service, attr='ip')}{self.service_get(service, attr='endpoint')}"

            if service in workers.swarm.services:
                command = COMMAND_STOP + COMMAND_RESTART
            else:
                command = COMMAND_START

            self.write_row(
                [
                    command,
                    f"{service}{self.service_get(service, attr='service_name',  default='')}",
                    f"{service in workers.swarm.services}",
                    f"{self.service_get(service, attr='node')}",
                    f"{self.service_get(service, attr='id')}",
                    f"{self.service_get(service, attr='container_id')}",
                    f"{self.service_get(service, attr='image')}",
                    url,
                ],
                row=row,
            )
        return row + 2

    # ----------------------------------------------------------------------
    def update_volumes_stats(self, name, start=0):
        """"""
        self.stdscr.addstr(start, 0, name)
        networks = sorted(
            [
                v
                for v in workers.swarm.client.volumes.list()
                if v.name.endswith('-volume')
            ],
            key=lambda s: s.name,
        )
        for row, volume in enumerate(networks, start=start + 1):
            self.stdscr.addstr(row, 0, volume.name)
        return row + 2

    # ----------------------------------------------------------------------
    def update_host_workers(self, name, start=0):
        """"""
        self.stdscr.addstr(start, 0, name)

        self.write_row(
            [name] + ['', 'PID', 'STATUS', 'RUN', 'ENV'], row=start, m=0
        )

        host_workers.load_ids()
        process = host_workers.ids.copy()

        if not process:
            return start + 2

        for row, service in enumerate(process, start=start + 1):

            self.row_host_workers[row] = service

            if process[service]['status'] in ['None', 'zombie']:
                command = COMMAND_START
            else:
                command = COMMAND_TERMINATE + COMMAND_RESTART

            self.write_row(
                [
                    command,
                    service,
                    str(process[service]['pid']),
                    process[service]['status'],
                    process[service]['run'],
                    str(process[service]['env']),
                ],
                row=row,
            )

        return row + 2


# ----------------------------------------------------------------------
def main():
    """"""
    while True:
        try:
            curses.wrapper(Stats)
            if not CONTINUE:
                break
        except Exception as e:
            pass
    sys.exit()


if __name__ == '__main__':
    main()



from foundation.radiant.utils import environ
import logging
import requests


########################################################################
class ChaskiLogging(logging.Handler):
    """"""

    # ----------------------------------------------------------------------
    def __init__(self):
        """"""
        super().__init__()

        self.topic = environ('WORKER_NAME', 'logs')
        formatter = logging.Formatter(f'%(levelname)s: {self.topic} [%(asctime)s]: %(message)s')
        self.setFormatter(formatter)

    # ----------------------------------------------------------------------
    def emit(self, record):
        """"""
        try:
            log_message = self.format(record)
            params = {
                "topic": self.topic,
                "message": log_message,
            }
            requests.get("http://chaski-api-logger-worker:51115/", params=params)

        except Exception as e:
            logging.error(f"Error enviando el log a Chaski: {e}")


try:
    custom_handler = ChaskiLogging()
    custom_handler.setLevel(logging.INFO)
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    root_logger.addHandler(custom_handler)
except Exception as e:
    logging.warning('Impossible to connect logging with Chaski')
    logging.warning(str(e))



#!/usr/bin/env python

from colorama import init
from colorama import Fore, Back
from chaski.streamer import ChaskiStreamer
import argparse
import logging
# from datetime import datetime

import asyncio

from foundation.utils import Workers

init(autoreset=True)

parser = argparse.ArgumentParser(description="Start an HCI logger.")
parser.add_argument(
    '-t',
    '--topic',
    action='append',
    default=['ALL'],
    help="Topics to be logged.",
)
parser.add_argument(
    '-l', '--loglevel', default='NOTSET', help="Set the logging level"
)
parser.add_argument(
    '-a', '--advertise_addr', default=None, help="Advertise address."
)

args = parser.parse_args()

logging.basicConfig()

workers = Workers(swarm_advertise_addr=args.advertise_addr)


# ----------------------------------------------------------------------
async def run():
    """"""

    consumer = ChaskiStreamer(
        subscriptions=['logs'], run=False
    )

    await consumer.connect('127.0.0.1', port=51114)
    asyncio.create_task(consumer.run())

    if args.topic == ['ALL']:
        topics = list(
            filter(
                lambda topic: not topic.startswith('__'),
                consumer.subscriptions,
            )
        )
        topics = topics + workers.swarm.services
    else:
        topics = args.topic

    if topics:
        consumer.subscribe(topics)
    print(Back.CYAN + Fore.BLACK + f'Topics: {", ".join(topics)}')

    async with consumer as message_queue:
        async for msg in message_queue:

            if not msg:
                continue

            if msg.topic == 'logs':
                print(msg.data)
                continue


            # print('+'+msg.data)

            msg = f'{msg.data.replace("@", ": ")}'

            log_level = getattr(logging, args.loglevel)
            msg_level = getattr(logging, msg.split(':')[0], -1)

            # print('+'+msg + f" <{msg_level}-{log_level}>")

            if msg_level < log_level:
                continue

            if msg.startswith('CRITICAL'):
                print(Back.RED + msg)
            elif msg.startswith('ERROR'):
                print(Fore.RED + msg)
            elif msg.startswith('WARNING'):
                print(Fore.YELLOW + msg)
            elif msg.startswith('INFO'):
                print(Fore.GREEN + msg)
            elif msg.startswith('DEBUG'):
                print(Fore.BLUE + msg)
            else:
                print(msg)


# ----------------------------------------------------------------------
def main():
    """"""
    asyncio.run(run())


if __name__ == '__main__':
    main()



from foundation.radiant.server import FrameworkAPI, render
from browser import html, document

# Material 3
# https://m3.material.io/develop/web
import material_3 as md


########################################################################
class BareMinimum(FrameworkAPI):

    # ----------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)
        self.add_css_file('styles/styles.css')

        # md_list = md.list()
        # self.body <= md_list
        # for i, service in enumerate(self.MyClass.services()):
            # list_item = md.list_item(headline=service, active=(i == 0))
            # md_list <= list_item

        self.body <= self.main_area()

        md_list = md.list()
        document.select_one('#radiant-drawer') <= md_list
        for i in range(10):
            list_item = md.list_item(headline=f'Item {i}', active=(i == 0))
            md_list <= list_item

        # button = md.text_button('Logging')
        # button.bind('click', self.send_log)
        # self.body <= button

        document.select_one('#radiant-main_area') <= html.PRE(self.swarm.get_join_command())

    # # ----------------------------------------------------------------------
    # def send_log(self, evt):
        # """"""
        # print('logging message')
        # self.logging.warning('logging message')

    # ----------------------------------------------------------------------
    def main_area(self):
        """"""
        return render('templates/main_area.html', context={})


if __name__ == '__main__':
    BareMinimum(logs=False)



import docker
import logging
import netifaces


########################################################################
class Swarm:
    """"""

    # ----------------------------------------------------------------------
    def __init__(
        self, base_url='unix://var/run/docker.sock', advertise_addr=''
    ):
        """Constructor"""
        self.client = docker.DockerClient(base_url=base_url)
        try:
            logging.warning("Starting swarm...")
            if not self.client.swarm.attrs.get('ID'):
                swarm_init_response = self.client.swarm.init(
                    advertise_addr=advertise_addr
                )
                logging.warning(f"Swarm started, ID: {swarm_init_response}")
            else:
                logging.warning(
                    f"Swarm is already running, ID: {self.client.swarm.attrs.get('ID')}"
                )
        except docker.errors.APIError as e:
            logging.warning("Error:", str(e))
            return

        self.create_networks()

    # ----------------------------------------------------------------------
    def create_networks(self):
        """"""
        self.networks = ['hci_network']
        logging.warning("Creating networks...")
        for network in self.networks:
            if network not in [n.name for n in self.client.networks.list()]:
                self.client.networks.create(network, driver="overlay")
                logging.warning(f"Created network '{network}'")

    # ----------------------------------------------------------------------
    def create_volume(self, volume_name):
        """"""
        logging.warning("Creating volumes...")
        if not volume_name in self.volumes:
            if not volume_name.endswith('-volume'):
                volume_name = f'{volume_name}-volume'
            self.client.volumes.create(name=volume_name)
        else:
            logging.warning(f"Volume '{volume_name}' already exists")

        return volume_name

    # ----------------------------------------------------------------------
    def delete_volume(self, volume_name):
        """"""
        logging.warning("Deleting volumes...")
        try:
            # Fetch the volume
            volume = self.client.volumes.get(volume_name)
            # Remove the volume
            volume.remove()
            logging.info(f"Volume '{volume_name}' has been deleted.")
        except docker.errors.NotFound:
            logging.warning(f"Volume '{volume_name}' does not exist.")
        except docker.errors.APIError as e:
            logging.error(
                f"An error occurred while attempting to remove the volume: {e}"
            )

        return None

    # ----------------------------------------------------------------------
    @property
    def services(self, attr='name'):
        """"""
        return [
            getattr(service, attr) for service in self.client.services.list()
        ]

    # ----------------------------------------------------------------------
    @property
    def containers(self, attr='id'):
        """"""
        return [
            getattr(container, attr)
            for container in self.client.containers.list()
        ]

    # ----------------------------------------------------------------------
    @property
    def volumes(self):
        """"""
        return [
            v.name
            for v in self.client.volumes.list()
            if v.name.endswith('-volume')
        ]

    # ----------------------------------------------------------------------
    def stop_service(self, service_name):
        """"""
        service = self.client.services.get(service_name)
        service.remove()

    # ----------------------------------------------------------------------
    def stop_all_services(self):
        """"""
        for service in self.services:
            self.stop_service(service)

    # ----------------------------------------------------------------------
    def stats(self, service_name):
        """"""
        service = self.client.services.get(service_name)
        stats = []
        for task in service.tasks():
            container_id = task['Status']['ContainerStatus']['ContainerID']
            if container_id in self.containers:
                stats.append(
                    self.client.containers.get(container_id).stats(
                        stream=False
                    )
                )
        return stats

    # ----------------------------------------------------------------------
    def start_ntp(
        self, service_name="ntp-service", port=123, restart=False, tag='1.0'
    ):
        """"""
        if restart and (service_name in self.services):
            self.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        service = self.client.services.create(
            image=f'dunderlab/ntp:{tag}',
            name=service_name,
            networks=self.networks,
            endpoint_spec={
                'Ports': [
                    {
                        'Protocol': 'udp',
                        'PublishedPort': port,
                        'TargetPort': 123,
                    },
                ]
            },
            env=[
                f"PORT={port}",
            ],
        )

        return service_name in self.services

    # ----------------------------------------------------------------------
    def start_jupyterlab(
        self,
        service_name="jupyterlab-service",
        port=8888,
        restart=False,
        tag='1.1',
        volume_name=None,
        mounts=None,
    ):
        """"""
        if restart and (service_name in self.services):
            self.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        if volume_name is None:
            volume_name = self.create_volume(service_name)
        else:
            volume_name = self.create_volume(volume_name)

        docker_mounts = []
        if mounts:
            for source, target in mounts:

                docker_mounts.append(
                    docker.types.Mount(
                        type='bind', source=source, target=target
                    )
                )

        service = self.client.services.create(
            image=f"dunderlab/python312:{tag}",
            name=service_name,
            networks=self.networks,
            command=[
                "/bin/bash",
                "-c",
                "jupyter lab --notebook-dir='/app' --ip=0.0.0.0 --port=8888 --allow-root --NotebookApp.token='' --NotebookApp.password=''",
            ],
            endpoint_spec={
                'Ports': [
                    {
                        'Protocol': 'tcp',
                        'PublishedPort': 8888,
                        'TargetPort': port,
                    },
                ]
            },
            hosts=self.extra_host(),
            mounts=[
                docker.types.Mount(
                    type='bind',
                    source='/var/run/docker.sock',
                    target='/var/run/docker.sock',
                ),
                docker.types.Mount(
                    target='/app',
                    source=volume_name,
                    type="volume",
                    read_only=False,
                ),
                *docker_mounts,
            ],
            env=[
                f"PORT={port}",
                # "NTP_SERVER=ntp-service",
            ],
        )
        return service_name in self.services

    # ----------------------------------------------------------------------
    def start_kafka(
        self,
        kafka_service_name="kafka-service",
        zookeeper_service_name="zookeeper-service",
        kafka_port=9092,
        kafka_port_external=19092,
        zookeeper_port=2181,
        restart=False,
        tag='1.1',
    ):
        """"""
        if restart and (kafka_service_name in self.services):
            self.stop_service(kafka_service_name)
            logging.warning(f"Restarting service '{kafka_service_name}'")

        if restart and (zookeeper_service_name in self.services):
            self.stop_service(zookeeper_service_name)
            logging.warning(f"Restarting service '{zookeeper_service_name}'")

        if not kafka_service_name in self.services:
            kafka_service = self.client.services.create(
                image=f"dunderlab/kafka:{tag}",
                # restart_policy=docker.types.RestartPolicy(condition='any'),
                name=kafka_service_name,
                networks=self.networks,
                endpoint_spec={
                    'Ports': [
                        {
                            'Protocol': 'tcp',
                            'PublishedPort': kafka_port,
                            'TargetPort': kafka_port,
                        },
                        {
                            'Protocol': 'tcp',
                            'PublishedPort': kafka_port_external,
                            'TargetPort': kafka_port_external,
                        },
                    ]
                },
                env=[
                    f"KAFKA_ZOOKEEPER_CONNECT={zookeeper_service_name}:{zookeeper_port}",
                    f"KAFKA_LISTENER_SECURITY_PROTOCOL_MAP=PLAINTEXT:PLAINTEXT,PLAINTEXT_HOST:PLAINTEXT",
                    f"KAFKA_ADVERTISED_LISTENERS=PLAINTEXT://{kafka_service_name}:{kafka_port},PLAINTEXT_HOST://localhost:{kafka_port_external}",
                    f"PORT={kafka_port_external}",
                ],
            )
        else:
            logging.warning(f"Service '{kafka_service_name}' already exist")

        if not zookeeper_service_name in self.services:
            zookeeper_service = self.client.services.create(
                image=f"dunderlab/zookeeper:{tag}",
                name=zookeeper_service_name,
                networks=self.networks,
                endpoint_spec={
                    'Ports': [
                        {
                            'Protocol': 'tcp',
                            'PublishedPort': zookeeper_port,
                            'TargetPort': zookeeper_port,
                        },
                    ]
                },
                env=[
                    f"ZOOKEEPER_CLIENT_PORT={zookeeper_port}",
                ],
            )
        else:
            logging.warning(
                f"Service '{zookeeper_service_name}' already exist"
            )

        return (
            kafka_service_name in self.services,
            zookeeper_service_name in self.services,
        )

    # ----------------------------------------------------------------------
    def start_kafka_logs(
        self,
        kafka_service_name="kafka-logs-service",
        zookeeper_service_name="zookeeper-logs-service",
        kafka_port=9093,
        kafka_port_external=19093,
        zookeeper_port=2182,
        restart=False,
        tag='1.1',
    ):
        """"""
        return self.start_kafka(
            kafka_service_name,
            zookeeper_service_name,
            kafka_port,
            kafka_port_external,
            zookeeper_port,
            restart,
            tag,
        )

    # ----------------------------------------------------------------------
    def start_timescaledb(
        self,
        service_name="timescaledb-service",
        port=5432,
        volume_name=None,
        restart=False,
        tag='latest-pg15',
    ):
        """"""
        if restart and (service_name in self.services):
            self.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        if volume_name is None:
            volume_name = self.create_volume(service_name)
        else:
            volume_name = self.create_volume(volume_name)

        timescaledb_service = self.client.services.create(
            image=f"timescale/timescaledb:{tag}",
            name=service_name,
            networks=self.networks,
            env=[
                "POSTGRES_PASSWORD=password",
                "POSTGRES_USER=postgres",
                "POSTGRES_DB=timescaledb",
                "POSTGRES_MAX_CONNECTIONS=500",
                f"PORT={port}",
            ],
            endpoint_spec={
                'Ports': [
                    {
                        'Protocol': 'tcp',
                        'PublishedPort': 5432,
                        'TargetPort': port,
                    },
                ]
            },
            mounts=[
                docker.types.Mount(
                    target='/var/lib/postgresql/data',
                    source=volume_name,
                    type="volume",
                    read_only=False,
                ),
            ],
        )
        return service_name in self.services

    # ----------------------------------------------------------------------
    def get_join_command(self):
        """"""
        swarm_info = self.client.info().get('Swarm')
        if swarm_info and swarm_info.get('ControlAvailable'):
            worker_join_token = self.client.swarm.attrs['JoinTokens'][
                'Worker'
            ]
            manager_addr = swarm_info.get('RemoteManagers')[0].get('Addr')
            return f'docker swarm join --token {worker_join_token} {manager_addr}'

    # ----------------------------------------------------------------------
    def advertise_addr(self):
        """"""
        swarm_info = self.client.info().get('Swarm')
        if swarm_info and swarm_info.get('ControlAvailable'):
            manager_addr = swarm_info.get('RemoteManagers')[0].get('Addr')
            return manager_addr

    # ----------------------------------------------------------------------
    def extra_host(self):
        """"""
        interfaces = netifaces.interfaces()
        ips = {}

        for interface in interfaces:
            addrs = netifaces.ifaddresses(interface)
            if netifaces.AF_INET in addrs:
                ip_info = addrs[netifaces.AF_INET][0]
                ips[interface] = ip_info['addr']

        return {
            f"host.docker.{interface}": ips[interface] for interface in ips
        }



import asyncio
import os
from chaski.remote import ChaskiRemote

# ----------------------------------------------------------------------
async def run(ip, port, name):
    """"""
    server = ChaskiRemote(
        ip=ip,
        port=port,
        name=name,
        available=os.getenv('modules', '').split(','),
        run=False,
    )

    print(f"CA Address: {server.address}")
    await server.run()


# ----------------------------------------------------------------------
def main(ip, port, name):
    """"""
    asyncio.run(run(ip, port, name))


main(ip='0.0.0.0', port=51112, name='Foundation Chaski Remote')



from chaski.streamer import ChaskiStreamer
from fastapi import FastAPI
import uvicorn
from contextlib import asynccontextmanager
import asyncio

app = FastAPI()

streamer = ChaskiStreamer(run=False)

@asynccontextmanager
async def lifespan(app: FastAPI):
    await streamer.connect('chaski-logger-root-worker', 51114)
    asyncio.create_task(streamer.run())
    yield
    # Shutdown code can go here

app = FastAPI(lifespan=lifespan)

@app.get("/")
async def log(topic, message):
    await streamer.push(topic, message)
    return {"ok": "True"}


import uvicorn
uvicorn.run(app, host="0.0.0.0", port=51115)



#!/usr/bin/env python

import os
import argparse
from foundation.utils import Workers
from foundation.utils.workers import WORKER_NAME

parser = argparse.ArgumentParser(description="Start an HCI worker.")
parser.add_argument(
    '-w', '--worker', default='.', help="Path of the worker's directory."
)
parser.add_argument(
    '-s', '--service_name', default=None, help="Name of the service."
)
parser.add_argument(
    '-r',
    '--restart',
    action='store_true',
    default=True,
    help="Option to restart the worker.",
)
parser.add_argument(
    '-k',
    '--kill',
    action='store_true',
    default=False,
    help="Option to kill the worker.",
)

args = parser.parse_args()

workers = Workers()


# ----------------------------------------------------------------------
def main():
    """"""

    if args.kill:
        workers.swarm.stop_service(WORKER_NAME.format(args.service_name))
    else:
        w = os.curdir if args.worker == '.' else args.worker
        workers.start_worker(
            w, service_name=args.service_name, restart=args.restart
        )


if __name__ == '__main__':
    main()



import os
import random
import socket
import logging

from string import digits, ascii_letters
import docker
from foundation.utils.swarm import Swarm
from foundation.workers import select_worker

WORKER_NAME = "{}-worker"


########################################################################
class Workers:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, swarm=None, swarm_advertise_addr=None):
        """Constructor"""

        if (swarm_advertise_addr) and (swarm is None):
            self.swarm = Swarm(advertise_addr=swarm_advertise_addr)
        elif (swarm_advertise_addr is None) and (swarm):
            self.swarm = swarm
        else:
            self.swarm = Swarm()

    # ----------------------------------------------------------------------
    def gen_worker_name(self, length=8):
        """"""
        id_ = "".join(
            [random.choice(ascii_letters + digits) for _ in range(length)]
        )
        if not WORKER_NAME.format(id_) in self.swarm.services:
            return WORKER_NAME.format(id_)
        return self.gen_worker_name(length)

    # ----------------------------------------------------------------------
    def get_open_port(
        self,
    ):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    # ----------------------------------------------------------------------
    def stop_all_workers(self):
        """"""
        for worker in self.swarm.services:
            if worker.endswith(WORKER_NAME.format("")):
                self.swarm.stop_service(worker)

    # ----------------------------------------------------------------------
    def start_django_worker(
        self,
        worker_path,
        service_name=None,
        port=None,
        restart=False,
        image='djangoship',
        tag=None,
        endpoint='',
        env={},
    ):
        """"""
        if tag == None and image == 'djangoship':
            tag = '1.4'
        if tag == None and image == 'djangorun':
            tag = '1.1'

        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
            app_name = os.path.split(worker_path)[-1]
        elif os.path.exists(select_worker(worker_path)):
            app_name = worker_path
            service_name = WORKER_NAME.format(worker_path.replace("_", "-"))
            worker_path = select_worker(worker_path)
        else:
            logging.warning(f"Service {worker_path} doesn't exits.")
            return

        if port is None:
            port = self.get_open_port()

        if service_name is None:
            service_name = self.gen_worker_name()
        else:
            service_name = WORKER_NAME.format(service_name)
        service_name_env = os.path.split(worker_path)[-1]

        if restart and (service_name in self.swarm.services):
            self.swarm.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.swarm.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        service = self.swarm.client.services.create(
            image=f"dunderlab/{image}:{tag}",
            name=service_name,
            networks=self.swarm.networks,
            endpoint_spec={
                'Ports': [
                    {
                        'Protocol': 'tcp',
                        'PublishedPort': port,
                        'TargetPort': 80,
                    },
                ]
            },
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=worker_path,
                    target=f"/app/{image}",
                    read_only=False,
                ),
                docker.types.Mount(
                    type="bind",
                    source="/var/run/docker.sock",
                    target="/var/run/docker.sock",
                ),
            ],
            env={
                "DJANGOPROJECT": app_name,
                "PORT": port,
                "ENDPOINT": endpoint,
                "SERVICE_NAME": service_name_env,
                "WORKER_NAME": service_name,
                **env,
            },
        )

        return port

    # ----------------------------------------------------------------------
    def start_brython_worker(
        self,
        worker_path,
        service_name=None,
        port=None,
        run="main.py",
        restart=False,
        tag='1.1',
        env={},
        request_ports={},
    ):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
        elif os.path.exists(select_worker(worker_path)):
            service_name = WORKER_NAME.format(worker_path.replace("_", "-"))
            worker_path = select_worker(worker_path)

        if port is None:
            port = self.get_open_port()

        if service_name is None:
            service_name = self.gen_worker_name()
        else:
            service_name = WORKER_NAME.format(service_name)
        service_name_env = os.path.split(worker_path)[-1]

        if restart and (service_name in self.swarm.services):
            self.swarm.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.swarm.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        endpoint_spec_ports = []
        env_ports = {}
        for port_env in request_ports:
            port_ = request_ports[port_env]
            if port_ is None:
                port_ = self.get_open_port()
            endpoint_spec_ports.append(
                {
                    'Protocol': 'tcp',
                    'PublishedPort': port_,
                    'TargetPort': port_,
                }
            )
            env_ports[port_env] = port_

        service = self.swarm.client.services.create(
            image=f"dunderlab/python312:{tag}",
            name=service_name,
            networks=self.swarm.networks,
            command=[
                "/bin/bash",
                "-c",
                # f"if [ -f \"/app/worker/requirements.txt\" ]; then pip install --root-user-action=ignore -r /app/worker/requirements.txt; fi && startup.sh && python /app/worker/{run}",
                f"ntpd -g && if [ -f \"/app/worker/requirements.txt\" ]; then pip install --root-user-action=ignore -r /app/worker/requirements.txt; fi && if [ -f \"/app/worker/startup.sh\" ]; then /app/worker/startup.sh; fi && python /app/worker/{run}",
            ],
            endpoint_spec={
                'Ports': [
                    {
                        'Protocol': 'tcp',
                        'PublishedPort': port,
                        'TargetPort': port,
                    },
                    *endpoint_spec_ports,
                ]
            },
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=worker_path,
                    target="/app/worker",
                    read_only=False,
                ),
                docker.types.Mount(
                    type="bind",
                    source="/var/run/docker.sock",
                    target="/var/run/docker.sock",
                ),
            ],
            env={
                # "STREAM": port_stream,
                "RADIANT": port,
                "PORT": port,
                "SERVICE_NAME": service_name_env,
                "WORKER_NAME": service_name,
                **env_ports,
                **env,
            },
        )

        return port

    # ----------------------------------------------------------------------
    def start_python_worker(
        self,
        worker_path,
        service_name=None,
        port=None,
        run="main.py",
        restart=False,
        tag='1.1',
        env={},
    ):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
        elif os.path.exists(select_worker(worker_path)):
            service_name = WORKER_NAME.format(worker_path.replace("_", "-"))
            worker_path = select_worker(worker_path)

        if port is None:
            port = self.get_open_port()

        if service_name is None:
            service_name = self.gen_worker_name()
        else:
            service_name = WORKER_NAME.format(service_name)
        service_name_env = os.path.split(worker_path)[-1]

        if restart and (service_name in self.swarm.services):
            self.swarm.stop_service(service_name)
            logging.warning(f"Restarting service '{service_name}'")
        elif service_name in self.swarm.services:
            logging.warning(f"Service '{service_name}' already exist")
            return

        service = self.swarm.client.services.create(
            image=f"dunderlab/python312:{tag}",
            name=service_name,
            networks=self.swarm.networks,
            command=[
                "/bin/bash",
                "-c",
                f"ntpd -g && if [ -f \"/app/worker/requirements.txt\" ]; then pip install --root-user-action=ignore -r /app/worker/requirements.txt; fi && if [ -f \"/app/worker/startup.sh\" ]; then /app/worker/startup.sh; fi && python /app/worker/{run}",
                # f"if [ -f \"/app/worker/requirements.txt\" ]; then pip install --root-user-action=ignore -r /app/worker/requirements.txt; fi && startup.sh && python /app/worker/{run}",
            ],
            endpoint_spec={
                'Ports': [
                    {
                        'Protocol': 'tcp',
                        'PublishedPort': port,
                        'TargetPort': port,
                    },
                ]
            },
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=worker_path,
                    target="/app/worker",
                    read_only=False,
                ),
                docker.types.Mount(
                    type="bind",
                    source="/var/run/docker.sock",
                    target="/var/run/docker.sock",
                ),
            ],
            env={
                "PORT": port,
                "SERVICE_NAME": service_name_env,
                "WORKER_NAME": service_name,
                **env,
            },
        )

        return port

    # ----------------------------------------------------------------------
    def start_worker(self, worker_path, **kwargs):
        """"""
        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)
        elif os.path.exists(select_worker(worker_path)):
            if not 'service_name' in kwargs:
                kwargs['service_name'] = worker_path
            worker_path = select_worker(worker_path)
            kwargs['service_name'] = kwargs['service_name'].replace('_', '-')

        if os.path.exists(os.path.join(worker_path, 'manage.py')):
            logging.warning('Running a Django worker')
            return self.start_django_worker(worker_path, **kwargs)
        elif os.path.exists(os.path.join(worker_path, 'main.py')):
            with open(os.path.join(worker_path, 'main.py'), 'r') as file:
                content = file.read()
                if 'from foundation.radiant.server' in content:
                    logging.warning('Running a Brython worker')
                    return self.start_brython_worker(worker_path, **kwargs)
                else:
                    logging.warning('Running a Pyhton worker')
                    return self.start_python_worker(worker_path, **kwargs)
        else:
            logging.warning('Impossible to detect a Worker')



import os
import subprocess
import pickle
import random
from string import digits, ascii_letters
import psutil

user_home = os.path.expanduser('~')
project_folder = os.path.join(user_home, 'foundation')
os.makedirs(project_folder, exist_ok=True)

WORKER_NAME = "{}-host-worker"


########################################################################
class HostWorker:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, env='python3'):
        """"""
        self.log_file = os.path.join(project_folder, 'process_info')
        self.env = env

        try:
            self.load_ids(drop=True)
        except (FileNotFoundError, EOFError):
            self.ids = {}
            self.save_ids()

    # ----------------------------------------------------------------------
    def load_ids(self, drop=False):
        """"""
        with open(self.log_file, 'rb') as f:
            self.ids = pickle.load(f)

        self.update_status(drop)

    # ----------------------------------------------------------------------
    def save_ids(self):
        """"""
        with open(self.log_file, 'wb') as f:
            pickle.dump(self.ids, f)

    # ----------------------------------------------------------------------
    def start_worker(
        self,
        worker_path,
        service_name=None,
        run="main.py",
        restart=False,
        tag='1.1',
        env={},
    ):
        """"""

        if os.path.isabs(worker_path) or os.path.exists(worker_path):
            worker_path = os.path.abspath(worker_path)

        if service_name is None:
            service_name = WORKER_NAME.format(
                os.path.split(worker_path)[-1].replace("_", "-")
            )

        service_name_env = service_name.replace('-host-worker', '')

        if service_name in self.ids:
            if restart:
                self.restart(service_name)
            return

        with open(os.path.join(worker_path, 'logfile'), 'w') as logfile:
            process = subprocess.Popen(
                [self.env, os.path.join(worker_path, run)],
                env={
                    "SERVICE_NAME": service_name_env,
                    "WORKER_NAME": service_name,
                    **env,
                },
                stdout=logfile,
                stderr=logfile,
                preexec_fn=os.setpgrp,
            )

        self.ids[service_name] = {
            'pid': process.pid,
            'restart': restart,
            'service_name_env': service_name_env,
            'worker_path': worker_path,
            'run': run,
            'env': env,
        }

        self.save_ids()

    # ----------------------------------------------------------------------
    def stop(self, service_name, command='kill'):
        """"""
        pid = self.ids[service_name]['pid']

        try:
            process = psutil.Process(pid)
            getattr(process, command)()

        except psutil.NoSuchProcess:
            pass

    # ----------------------------------------------------------------------
    def restart(self, service_name):
        """"""
        self.stop(service_name)

        proc = self.ids[service_name]
        with open(
            os.path.join(proc['worker_path'], 'logfile'), 'w'
        ) as logfile:
            process = subprocess.Popen(
                [self.env, os.path.join(proc['worker_path'], proc['run'])],
                env={
                    "SERVICE_NAME": proc['service_name_env'],
                    "WORKER_NAME": service_name,
                    **proc['env'],
                },
                stdout=logfile,
                stderr=logfile,
                preexec_fn=os.setpgrp,
            )

        proc['pid'] = process.pid
        self.save_ids()

    # ----------------------------------------------------------------------
    def gen_worker_name(self, length=8):
        """"""
        id_ = "".join(
            [random.choice(ascii_letters + digits) for _ in range(length)]
        )
        if not WORKER_NAME.format(id_) in self.ids:
            return WORKER_NAME.format(id_)

        return self.gen_worker_name(length)

    # ----------------------------------------------------------------------
    def update_status(self, drop):
        """"""
        for name in self.ids:
            pid = self.ids[name]['pid']
            try:
                process = psutil.Process(pid)
                status = process.status()

                self.ids[name]['status'] = status
            except psutil.NoSuchProcess:
                self.ids[name]['status'] = 'None'

        if drop:
            ids = self.ids.copy()
            for name in self.ids:
                if self.ids[name]['status'] == 'None':
                    ids.pop(name)
            self.ids = ids

        self.save_ids()



import pickle

from confluent_kafka import Consumer, KafkaError

from dunderlab.api.utils import JSON
from dunderlab.api import aioAPI as API
from dunderlab.api.utils import get_data

conf = {
    'bootstrap.servers': 'kafka-logs-service:9093',
    'group.id': 'api',
    'auto.offset.reset': 'latest',
}

TOKEN = {
    "refresh": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ0b2tlbl90eXBlIjoicmVmcmVzaCIsImV4cCI6MTcyNTIyODA1OSwiaWF0IjoxNjkzNjkyMDU5LCJqdGkiOiIyMjg3Zjc2MTk5MDY0MzJiYTc0ZTZiMWQ4NWNiYzBiNCIsInVzZXJfaWQiOjF9.6wq-irScHTJ8nf8U0ruByrrYiEvEkqAbZFKTyF_A8Yc",
    "access": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ0b2tlbl90eXBlIjoiYWNjZXNzIiwiZXhwIjoxNjkzNzc4NDU5LCJpYXQiOjE2OTM2OTIwNTksImp0aSI6ImRhZDgzNTU3NTk3MDQ3YzA5NDZhZWQ1OTczZmU5OWNkIiwidXNlcl9pZCI6MX0.RcO7-ms8tj8F-9JkOzfACYMhtByi7BKspOla7bhpAK0"
}

api = API('http://timescaledb-api-worker/timescaledbapp/', token=TOKEN['access'])

consumer = Consumer(conf)
consumer.subscribe(['timescaledb-api'])


# ----------------------------------------------------------------------
def write_on_api(data):
    """"""
    global api

    endpoint = data['endpoint']
    params = data['params']
    await getattr(api, endpoint).post(params)


try:
    while True:
        msg = consumer.poll(timeout=1)
        if msg is None:
            continue
        if msg.error():
            if msg.error().code() == KafkaError._PARTITION_EOF:
                print(f'Final del stream en el topic/partición: {msg.topic()}/{msg.partition()}')
            else:
                print(f'Error al recibir el mensaje: {msg.error()}')
        else:
            data = pickle.loads(msg.value())
            write_on_api(data)
except KeyboardInterrupt:
    pass
finally:
    consumer.close()
